# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""All-in-one public API module.

It's OK to use anything in anaconda_project as well, but anything
with an underscore prefix or inside ``anaconda_project.internal``
is considered private.

In this file, we try to export the interesting high-level
operations in one place so they are easy to find and import.  It
is a redundant but hopefully convenient wrapper around the entire
API.

"""
from __future__ import absolute_import

# This file shouldn't import anaconda_project.internal, because it's
# supposed to wrap other public API, not be the only public API.
from anaconda_project import prepare, project, provide, project_ops


class AnacondaProject(object):
    """Class containing a consolidated public API for convenience."""
    def __init__(self):
        """Construct an API instance."""
        pass

    def load_project(self, directory_path, frontend):
        """Load a project from the given directory.

        If there's a problem, the returned Project instance will
        have a non-empty ``problems`` attribute. So check
        ``project.problems`` when you get the result.
        ``project.problems`` can change anytime changes are made
        to a project; code must always be ready for a project to
        have some problems.

        Args:
            directory_path (str): path to the project directory
            frontend (Frontend): UX abstraction

        Returns:
            a Project instance

        """
        return project.Project(directory_path=directory_path, frontend=frontend)

    def create_project(self, directory_path, make_directory=False, name=None, icon=None, description=None):
        """Create a project skeleton in the given directory.

        Returns a Project instance even if creation fails or the directory
        doesn't exist, but in those cases the ``problems`` attribute
        of the Project will describe the problem.

        If the anaconda-project.yml already exists, this simply loads it.

        This will not prepare the project (create environments, etc.),
        use the separate prepare calls if you want to do that.

        Args:
            directory_path (str): directory to contain anaconda-project.yml
            make_directory (bool): True to create the directory if it doesn't exist
            name (str): Name of the new project or None to leave unset (uses directory name)
            icon (str): Icon for the new project or None to leave unset (uses no icon)
            description (str): Description for the new project or None to leave unset

        Returns:
            a Project instance
        """
        return project_ops.create(directory_path=directory_path,
                                  make_directory=make_directory,
                                  name=name,
                                  icon=icon,
                                  description=description)

    def prepare_project_locally(self,
                                project,
                                environ,
                                env_spec_name=None,
                                command_name=None,
                                command=None,
                                extra_command_args=None):
        """Prepare a project to run one of its commands.

        "Locally" means a machine where development will go on,
        contrasted with say a production deployment.

        This method takes any needed actions such as creating
        environments or starting services, without asking the user
        for permission.

        This method returns a result object. The result object has
        a ``failed`` property.  If the result is failed, the
        ``errors`` property has the errors.  If the result is not
        failed, the ``command_exec_info`` property has the stuff
        you need to run the project's default command, and the
        ``environ`` property has the updated environment. The
        passed-in ``environ`` is not modified in-place.

        You can update your original environment with
        ``result.update_environ()`` if you like, but it's probably
        a bad idea to modify ``os.environ`` in that way because
        the calling app won't want to have the project
        environment.

        The ``environ`` should usually be kept between
        preparations, starting out as ``os.environ`` but then
        being modified by the user.

        If the project has a non-empty ``problems`` attribute,
        this function returns the project problems inside a failed
        result. So ``project.problems`` does not need to be checked in
        advance.

        Args:
            project (Project): from the ``load_project`` method
            environ (dict): os.environ or the previously-prepared environ; not modified in-place
            env_spec_name (str): the package set name to require, or None for default
            command_name (str): which named command to choose from the project, None for default
            command (ProjectCommand): a command object (alternative to command_name)
            extra_command_args (list): extra args to include in the returned command argv

        Returns:
            a ``PrepareResult`` instance, which has a ``failed`` flag

        """
        return prepare.prepare_without_interaction(project=project,
                                                   environ=environ,
                                                   mode=provide.PROVIDE_MODE_DEVELOPMENT,
                                                   env_spec_name=env_spec_name,
                                                   command_name=command_name,
                                                   command=command,
                                                   extra_command_args=extra_command_args)

    def prepare_project_production(self,
                                   project,
                                   environ,
                                   env_spec_name=None,
                                   command_name=None,
                                   command=None,
                                   extra_command_args=None):
        """Prepare a project to run one of its commands.

        "Production" means some sort of production deployment, so
        services have to be 'real' and not some kind of
        local/temporary throwaway. We won't just start things up
        willy-nilly.

        We still do some things automatically in production
        though, such as creating environments.

        This method does not interact with the user; it "does the
        right thing" without asking.

        See ``prepare_project_locally()`` for additional details
        that also apply to this method.

        Args:
            project (Project): from the ``load_project`` method
            environ (dict): os.environ or the previously-prepared environ; not modified in-place
            env_spec_name (str): the package set name to require, or None for default
            command_name (str): which named command to choose from the project, None for default
            command (ProjectCommand): a command object (alternative to command_name)
            extra_command_args (list): extra args to include in the returned command argv

        Returns:
            a ``PrepareResult`` instance, which has a ``failed`` flag

        """
        return prepare.prepare_without_interaction(project=project,
                                                   environ=environ,
                                                   mode=provide.PROVIDE_MODE_PRODUCTION,
                                                   env_spec_name=env_spec_name,
                                                   command_name=command_name,
                                                   command=command,
                                                   extra_command_args=extra_command_args)

    def prepare_project_check(self,
                              project,
                              environ,
                              env_spec_name=None,
                              command_name=None,
                              command=None,
                              extra_command_args=None):
        """Prepare a project to run one of its commands.

        This version only checks the status of the project's
        requirements, but doesn't take any actions; it won't
        create files or start processes or anything like that.  If
        it returns a successful result, the project can be
        prepared without taking any further action.

        See ``prepare_project_locally()`` for additional details
        that also apply to this method.

        Args:
            project (Project): from the ``load_project`` method
            environ (dict): os.environ or the previously-prepared environ; not modified in-place
            env_spec_name (str): the package set name to require, or None for default
            command_name (str): which named command to choose from the project, None for default
            command (ProjectCommand): a command object (alternative to command_name)
            extra_command_args (list): extra args to include in the returned command argv

        Returns:
            a ``PrepareResult`` instance, which has a ``failed`` flag

        """
        return prepare.prepare_without_interaction(project=project,
                                                   environ=environ,
                                                   mode=provide.PROVIDE_MODE_CHECK,
                                                   env_spec_name=env_spec_name,
                                                   command_name=command_name,
                                                   command=command,
                                                   extra_command_args=extra_command_args)

    def unprepare(self, project, prepare_result, whitelist=None):
        """Attempt to clean up project-scoped resources allocated by prepare().

        This will retain any user configuration choices about how to
        provide requirements, but it stops project-scoped services.
        Global system services or other services potentially shared
        among projects will not be stopped.

        To stop a single service, use ``whitelist=["SERVICE_VARIABLE"]``.

        Args:
            project (Project): the project
            prepare_result (PrepareResult): result from the previous prepare
            whitelist (iterable of str or type): ONLY call shutdown commands for the listed env vars' requirements

        """
        return prepare.unprepare(project=project, prepare_result=prepare_result, whitelist=whitelist)

    def set_properties(self, project, name=None, icon=None, description=None):
        """Set simple properties on a project.

        This doesn't support properties which require prepare()
        actions to check their effects; see other calls such as
        ``add_packages()`` for those.

        This will fail if project.problems is non-empty.

        Args:
            project (``Project``): the project instance
            name (str): Name of the project or None to leave unmodified
            icon (str): Icon for the project or None to leave unmodified
            description (str): description for the project or None to leave unmodified

        Returns:
            a ``Status`` instance indicating success or failure
        """
        return project_ops.set_properties(project=project, name=name, icon=icon, description=description)

    def add_variables(self, project, env_spec_name, vars_to_add, defaults):
        """Add variables in anaconda-project.yml, optionally setting their defaults.

        Returns a ``Status`` instance which evaluates to True on
        success and has an ``errors`` property (with a list of error
        strings) on failure.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            vars_to_add (list of str): variable names
            defaults (dict): dictionary from keys to defaults, can be empty

        Returns:
            ``Status`` instance
        """
        return project_ops.add_variables(project=project,
                                         env_spec_name=env_spec_name,
                                         vars_to_add=vars_to_add,
                                         defaults=defaults)

    def remove_variables(self, project, env_spec_name, vars_to_remove, prepare_result=None):
        """Remove variables from anaconda-project.yml and unset their values in local project state.

        Returns a ``Status`` instance which evaluates to True on
        success and has an ``errors`` property (with a list of error
        strings) on failure.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            vars_to_remove (list of tuple): key-value pairs
            prepare_result (PrepareResult): result of a previous prepare or None

        Returns:
            ``Status`` instance
        """
        return project_ops.remove_variables(project=project,
                                            env_spec_name=env_spec_name,
                                            vars_to_remove=vars_to_remove,
                                            prepare_result=prepare_result)

    def set_variables(self, project, env_spec_name, vars_and_values, prepare_result=None):
        """Set variables' values in anaconda-project-local.yml.

        Returns a ``Status`` instance which evaluates to True on
        success and has an ``errors`` property (with a list of error
        strings) on failure.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            vars_and_values (list of tuple): key-value pairs
            prepare_result (PrepareResult): result of a previous prepare or None

        Returns:
            ``Status`` instance
        """
        return project_ops.set_variables(project=project,
                                         env_spec_name=env_spec_name,
                                         vars_and_values=vars_and_values,
                                         prepare_result=prepare_result)

    def unset_variables(self, project, env_spec_name, vars_to_unset, prepare_result=None):
        """Unset variables' values in anaconda-project-local.yml.

        Returns a ``Status`` instance which evaluates to True on
        success and has an ``errors`` property (with a list of error
        strings) on failure.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            vars_to_unset (list of str): variable names
            prepare_result (PrepareResult): result of a previous prepare or None

        Returns:
            ``Status`` instance
        """
        return project_ops.unset_variables(project=project,
                                           env_spec_name=env_spec_name,
                                           vars_to_unset=vars_to_unset,
                                           prepare_result=prepare_result)

    def add_download(self, project, env_spec_name, env_var, url, filename=None, hash_algorithm=None, hash_value=None):
        """Attempt to download the URL; if successful, add it as a download to the project.

        The returned ``Status`` should be a ``RequirementStatus`` for
        the download requirement if it evaluates to True (on success),
        but may be another subtype of ``Status`` on failure. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            env_var (str): env var to store the local filename
            url (str): url to download
            filename (optional, str): Name to give file or directory after downloading
            hash_algorithm (optional, str): Name of the algorithm to use for checksum verification
                                       must be present if hash_value is entered
            hash_value (optional, str): Checksum value to use for verification
                                           must be present if hash_algorithm is entered
        Returns:
            ``Status`` instance
        """
        return project_ops.add_download(project=project,
                                        env_spec_name=env_spec_name,
                                        env_var=env_var,
                                        url=url,
                                        filename=filename,
                                        hash_algorithm=hash_algorithm,
                                        hash_value=hash_value)

    def remove_download(self, project, env_spec_name, env_var, prepare_result=None):
        """Remove file or directory referenced by ``env_var`` from file system and the project.

        The returned ``Status`` will be an instance of ``SimpleStatus``. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            env_var (str): env var to store the local filename
            prepare_result (PrepareResult): result of a previous prepare

        Returns:
            ``Status`` instance
        """
        return project_ops.remove_download(project=project,
                                           env_spec_name=env_spec_name,
                                           env_var=env_var,
                                           prepare_result=prepare_result)

    def add_env_spec(self, project, name, packages, channels):
        """Attempt to create the environment spec and add it to anaconda-project.yml.

        The returned ``Status`` will be an instance of ``SimpleStatus``. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            name (str): environment name
            packages (list of str): packages (with optional version info, as for conda install)
            channels (list of str): channels (as they should be passed to conda --channel)

        Returns:
            ``Status`` instance
        """
        return project_ops.add_env_spec(project=project, name=name, packages=packages, channels=channels)

    def remove_env_spec(self, project, name):
        """Remove the environment spec from project directory and remove from anaconda-project.yml.

        Returns a ``Status`` subtype (it won't be a
        ``RequirementStatus`` as with some other functions, just a
        plain status).

        Args:
            project (Project): the project
            name (str): environment name

        Returns:
            ``Status`` instance
        """
        return project_ops.remove_env_spec(project=project, name=name)

    def export_env_spec(self, project, name, filename):
        """Export the environment spec as an environment.yml-type file.

        Returns a ``Status`` subtype (it won't be a
        ``RequirementStatus`` as with some other functions, just a
        plain status).

        Args:
            project (Project): the project
            name (str): environment spec name
            filename (str): file to export to

        Returns:
            ``Status`` instance
        """
        return project_ops.export_env_spec(project=project, name=name, filename=filename)

    def add_packages(self, project, env_spec_name, packages, channels, pip=False):
        """Attempt to install packages then add them to anaconda-project.yml.

        If the environment spec name is None rather than an env
        name, packages are added in the global packages
        section (to all environments).

        The returned ``Status`` should be a ``RequirementStatus`` for
        the environment requirement if it evaluates to True (on success),
        but may be another subtype of ``Status`` on failure. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            packages (list of str): packages (with optional version info, as for conda install)
            channels (list of str): channels (as they should be passed to conda --channel)
            pip (bool): Flag to request packages to be installed with pip if True else use Conda.

        Returns:
            ``Status`` instance

        """
        return project_ops.add_packages(project=project,
                                        env_spec_name=env_spec_name,
                                        packages=packages,
                                        channels=channels,
                                        pip=pip)

    def remove_packages(self, project, env_spec_name, packages, pip):
        """Attempt to remove packages from an environment spec in anaconda-project.yml.

        If the environment spec name is None rather than an env
        name, packages are removed from the global
        packages section (from all environments).

        The returned ``Status`` should be a ``RequirementStatus`` for
        the environment requirement if it evaluates to True (on success),
        but may be another subtype of ``Status`` on failure. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment name or None for all environments
            packages (list of str): packages
            pip (bool): Flag to request packages to be removed with pip if True else use Conda.

        Returns:
            ``Status`` instance

        """
        return project_ops.remove_packages(project=project, env_spec_name=env_spec_name, packages=packages, pip=pip)

    def lock(self, project, env_spec_name):
        """Attempt to freeze dependency versions in anaconda-project-lock.yml.

        If the env_spec_name is None rather than a name,
        all env specs are frozen.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs

        Returns:
            ``Status`` instance
        """
        return project_ops.lock(project=project, env_spec_name=env_spec_name)

    def update(self, project, env_spec_name):
        """Attempt to update frozen dependency versions in anaconda-project-lock.yml.

        If the env_spec_name is None rather than a name,
        all env specs are updated.

        If an env is not locked, this updates the installed dependencies but
        doesn't change anything about project configuration (does not save
        the lock file).

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs

        Returns:
            ``Status`` instance
        """
        return project_ops.update(project=project, env_spec_name=env_spec_name)

    def unlock(self, project, env_spec_name):
        """Attempt to unfreeze dependency versions in anaconda-project-lock.yml.

        If the env_spec_name is None rather than a name,
        all env specs are unfrozen.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs

        Returns:
            ``Status`` instance
        """
        return project_ops.unlock(project=project, env_spec_name=env_spec_name)

    def add_platforms(self, project, env_spec_name, platforms):
        """Attempt to add platforms the project supports.

        If the env_spec_name is None rather than an env name,
        packages are added in the global platforms section (to
        all environment specs).

        The returned ``Status`` should be a ``RequirementStatus`` for
        the environment requirement if it evaluates to True (on success),
        but may be another subtype of ``Status`` on failure. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            platforms (list of str): platforms to add

        Returns:
            ``Status`` instance
        """
        return project_ops.add_platforms(project=project, env_spec_name=env_spec_name, platforms=platforms)

    def remove_platforms(self, project, env_spec_name, platforms):
        """Attempt to remove platforms the project supports.

        If the env_spec_name is None rather than an env name,
        packages are added in the global platforms section (to
        all environment specs).

        The returned ``Status`` should be a ``RequirementStatus`` for
        the environment requirement if it evaluates to True (on success),
        but may be another subtype of ``Status`` on failure. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            platforms (list of str): platforms to remove

        Returns:
            ``Status`` instance
        """
        return project_ops.remove_platforms(project=project, env_spec_name=env_spec_name, platforms=platforms)

    def add_command(self, project, name, command_type, command, env_spec_name=None, supports_http_options=None):
        """Add a command to anaconda-project.yml.

        Returns a ``Status`` subtype (it won't be a
        ``RequirementStatus`` as with some other functions, just a
        plain status).

        Args:
           project (Project): the project
           name (str): name of the command
           command_type (str): choice of `bokeh_app`, `notebook`, `unix` or `windows` command
           command (str): the command line or filename itself
           env_spec_name (str): env spec to use with this command
           supports_http_options (bool): whether command supports --anaconda-project-* http server options

        Returns:
           a ``Status`` instance

        """
        return project_ops.add_command(project=project,
                                       name=name,
                                       command_type=command_type,
                                       command=command,
                                       env_spec_name=env_spec_name,
                                       supports_http_options=supports_http_options)

    def update_command(self, project, name, command_type=None, command=None, new_name=None):
        """Update attributes of a command in anaconda-project.yml.

        Returns a ``Status`` subtype (it won't be a
        ``RequirementStatus`` as with some other functions, just a
        plain status).

        Args:
           project (Project): the project
           name (str): name of the command
           command_type (str or None): choice of `bokeh_app`, `notebook`, `unix` or `windows` command
           command (str or None): the command line or filename itself; command_type must also be specified
           new_name (str or None): a new name to reference the command

        Returns:
           a ``Status`` instance
        """
        return project_ops.update_command(project=project,
                                          name=name,
                                          command_type=command_type,
                                          command=command,
                                          new_name=new_name)

    def remove_command(self, project, name):
        """Remove a command from anaconda-project.yml.

        Returns a ``Status`` subtype (it won't be a
        ``RequirementStatus`` as with some other functions, just a
        plain status).

        Args:
           project (Project): the project
           name (string): name of the command to be removed

        Returns:
           a ``Status`` instance
        """
        return project_ops.remove_command(project=project, name=name)

    def add_service(self, project, env_spec_name, service_type, variable_name=None):
        """Add a service to anaconda-project.yml.

        The returned ``Status`` should be a ``RequirementStatus`` for
        the service requirement if it evaluates to True (on success),
        but may be another subtype of ``Status`` on failure. A False
        status will have an ``errors`` property with a list of error
        strings.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            service_type (str): which kind of service
            variable_name (str): environment variable name (None for default)

        Returns:
            ``Status`` instance
        """
        return project_ops.add_service(project=project,
                                       env_spec_name=env_spec_name,
                                       service_type=service_type,
                                       variable_name=variable_name)

    def remove_service(self, project, env_spec_name, variable_name, prepare_result=None):
        """Remove a service to anaconda-project.yml.

        Returns a ``Status`` instance which evaluates to True on
        success and has an ``errors`` property (with a list of error
        strings) on failure.

        Args:
            project (Project): the project
            env_spec_name (str): environment spec name or None for all environment specs
            variable_name (str): environment variable name for the service requirement
            prepare_result (PrepareResult): result of a previous prepare or None

        Returns:
            ``Status`` instance
        """
        return project_ops.remove_service(project=project,
                                          env_spec_name=env_spec_name,
                                          variable_name=variable_name,
                                          prepare_result=prepare_result)

    def clean(self, project, prepare_result):
        """Blow away auto-provided state for the project.

        This should not remove any potential "user data" such as
        anaconda-project-local.yml.

        Args:
            project (Project): the project instance
            prepare_result (PrepareResult): result of a previous prepare

        Returns:
            a ``Status`` instance

        """
        return project_ops.clean(project=project, prepare_result=prepare_result)

    def archive(self, project, filename, pack_envs=False):
        """Make an archive of the non-ignored files in the project.

        Args:
            project (``Project``): the project
            filename (str): name of a zip, tar.gz, or tar.bz2 archive file
            pack_envs (bool): Flag to include conda-packs of each env_spec in the archive

        Returns:
            a ``Status``, if failed has ``errors``
        """
        return project_ops.archive(project=project, filename=filename, pack_envs=pack_envs)

    def unarchive(self, filename, project_dir, parent_dir=None, frontend=None):
        """Unpack an archive of the project.

        The archive can be untrusted (we will safely defeat attempts
        to put evil links in it, for example), but this function
        doesn't load or validate the unpacked project.

        The target directory must not exist or it's an error.

        project_dir can be None to auto-choose one.

        If parent_dir is non-None, place the project_dir in it. This is most useful
        if project_dir is None.

        Args:
            filename (str): name of a zip, tar.gz, or tar.bz2 archive file
            project_dir (str): the directory to place the project inside
            parent_dir (str): directory to place project_dir within
            frontend (Frontend): frontend instance representing current UX

        Returns:
            a ``Status``, if failed has ``errors``, on success has ``project_dir`` property.

        """
        return project_ops.unarchive(filename=filename,
                                     project_dir=project_dir,
                                     parent_dir=parent_dir,
                                     frontend=frontend)

    def upload(self, project, private=None, site=None, username=None, token=None, suffix='.tar.bz2', log_level=None):
        """Upload the project to the Anaconda server.

        Args:
            project (``Project``): the project
            private (bool): make project private
            site (str): site alias from Anaconda config
            username (str): Anaconda username
            token (str): Anaconda auth token
            log_level (str): Anaconda log level

        Returns:
            a ``Status``, if failed has ``errors``
        """
        return project_ops.upload(project=project,
                                  private=private,
                                  site=site,
                                  username=username,
                                  token=token,
                                  suffix=suffix,
                                  log_level=log_level)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
import os
from os.path import join

from anaconda_project.test.environ_utils import minimal_environ
from anaconda_project.test.project_utils import project_no_dedicated_env
from anaconda_project.prepare import (prepare_without_interaction)
from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents_completing_project_file)
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project import project

from anaconda_project.plugins import CommandTemplate, ArgsTransformerTemplate


def test_prepare_plugin_command(monkeypatch, tmpdir):
    called_with = {}
    cmd_name = 'custom-cmd'
    assert hasattr(project.plugins_api, 'get_plugins')

    def get_plugins_mock(cmd_type):
        return {'valid_package_plugin': plugin_init_mock}

    class TestTransformer(ArgsTransformerTemplate):
        def add_args(self, results, args):
            return ['--show']

    class TestCmd(CommandTemplate):
        args_transformer_cls = TestTransformer
        command = cmd_name

        def choose_args_and_shell(self, environ, extra_args=None):
            assert extra_args is None or isinstance(extra_args, list)

            shell = False
            args = [self.command_with_conda_prefix, 'custom-sub-cmd', '--%s.TESTARG' % self.command]

            return args + extra_args, shell

    def plugin_init_mock(*args, **kws):
        called_with['args'] = args
        called_with['kws'] = kws
        return TestCmd(*args, **kws)

    def check(dirname):
        # do not use monkeypatch
        # since with_directory_contents_completing_project_file
        # monkeypatches zipfile as, that is use by entry_points
        project.plugins_api.get_plugins = get_plugins_mock

        _project = project_no_dedicated_env(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(_project, environ=environ, command_name='foo')

        cmd_path = join(os.environ['CONDA_PREFIX'], 'bin', cmd_name)
        expected = [cmd_path, 'custom-sub-cmd', '--%s.TESTARG' % cmd_name, '--show']
        assert result.errors == []
        assert result
        assert result.command_exec_info.args == expected

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
    foo:
       valid_package_plugin: foo.py
packages:
  - notebook
""",
            "foo.py": "# foo",
        }, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os

from anaconda_project.internal.test.tmpfile_utils import (with_file_contents, with_named_file_contents,
                                                          with_directory_contents)

from anaconda_project.env_spec import (EnvSpec, _load_environment_yml, _load_requirements_txt,
                                       _find_out_of_sync_importable_spec)

from anaconda_project.conda_manager import CondaLockSet


def test_load_environment_yml():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None
        assert spec.name == 'foo'
        assert spec.conda_packages == ('bar=1.0', 'baz')
        assert spec.pip_packages == ('pippy', 'poppy==2.0')
        assert spec.channels == ('channel1', 'channel2')

        assert spec.logical_hash == 'e91a2263df510c9b188b132b801ba53aa99cc407'

    with_file_contents(
        """
name: foo
dependencies:
  - bar=1.0
  - baz
  - pip:
    - pippy
    - poppy==2.0
channels:
  - channel1
  - channel2
    """, check)


def test_load_environment_yml_with_prefix():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None
        assert spec.name == 'foo'
        assert spec.conda_packages == ('bar=1.0', 'baz')
        assert spec.pip_packages == ('pippy', 'poppy==2.0')
        assert spec.channels == ('channel1', 'channel2')

        assert spec.logical_hash == 'e91a2263df510c9b188b132b801ba53aa99cc407'

    with_file_contents(
        """
prefix: /opt/foo
dependencies:
  - bar=1.0
  - baz
  - pip:
    - pippy
    - poppy==2.0
channels:
  - channel1
  - channel2
    """, check)


def test_load_environment_yml_no_name():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None
        assert spec.name == os.path.basename(filename)
        assert spec.conda_packages == ('bar=1.0', 'baz')
        assert spec.pip_packages == ('pippy', 'poppy==2.0')
        assert spec.channels == ('channel1', 'channel2')

        assert spec.logical_hash == 'e91a2263df510c9b188b132b801ba53aa99cc407'

    with_file_contents(
        """
dependencies:
  - bar=1.0
  - baz
  - pip:
    - pippy
    - poppy==2.0
channels:
  - channel1
  - channel2
    """, check)


def test_load_environment_yml_with_broken_sections():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None
        assert spec.name == 'foo'
        assert spec.conda_packages == ()
        assert spec.pip_packages == ()
        assert spec.channels == ()

    with_file_contents("""
name: foo
dependencies: 42
channels: 57
    """, check)


def test_load_environment_yml_with_broken_pip_section():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None
        assert spec.name == 'foo'
        assert spec.conda_packages == ()
        assert spec.pip_packages == ()
        assert spec.channels == ()

    with_file_contents("""
name: foo
dependencies:
 - pip: 42
channels: 57
    """, check)


def test_load_requirements_txt():
    def check(filename):
        spec = _load_requirements_txt(filename)

        assert spec is not None
        assert spec.name == 'default'
        assert spec.conda_packages == ()
        assert spec.channels == ()
        assert spec.pip_packages == ('MyApp', 'Framework==0.9.4', 'Library>=0.2',
                                     'svn+http://myrepo/svn/MyThing#egg=MyThing')
        assert spec.pip_package_names_set == set(('MyApp', 'Framework', 'Library', 'MyThing'))

        assert spec.logical_hash == '784ba385d4cd468756e3cbc57f33e97afdc38059'

    with_file_contents(
        """
MyApp
# Comment; this is a framework
Framework==0.9.4

  # blank line above this indented comment!
 Library>=0.2
-e svn+http://myrepo/svn/MyThing#egg=MyThing
--index-url http://example.com/private-pypi/
--find-links http://example.com/private-packages/
    """, check)


def test_load_recursive_requirements_txt():
    def check(dirname):
        spec = _load_requirements_txt(os.path.join(dirname, "requirements.txt"))

        assert spec is not None
        assert spec.name == 'default'
        assert spec.pip_packages == ('a', 'b', 'c', 'd')

    with_directory_contents(
        {
            "requirements.txt": """
a
b
-r more-requirements.txt
        """,
            "more-requirements.txt": """
c
d
"""
        }, check)


def test_find_in_sync_environment_yml():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None

        (desynced, name) = _find_out_of_sync_importable_spec([spec], os.path.dirname(filename))
        assert desynced is None
        assert name is None

    with_named_file_contents(
        "environment.yml", """
name: foo
dependencies:
  - bar=1.0
  - baz
  - pip:
    - pippy
    - poppy==2.0
channels:
  - channel1
  - channel2
    """, check)


def test_find_out_of_sync_environment_yml():
    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None

        changed = EnvSpec(name=spec.name,
                          conda_packages=spec.conda_packages[1:],
                          pip_packages=spec.pip_packages,
                          channels=spec.channels)

        (desynced, name) = _find_out_of_sync_importable_spec([changed], os.path.dirname(filename))
        assert desynced is not None
        assert desynced.logical_hash == spec.logical_hash
        assert name == os.path.basename(filename)

    with_named_file_contents(
        "environment.yaml", """
name: foo
dependencies:
  - bar=1.0
  - baz
  - pip:
    - pippy
    - poppy==2.0
channels:
  - channel1
  - channel2
    """, check)


def test_load_environment_yml_does_not_exist():
    spec = _load_environment_yml("nopenopenope")
    assert spec is None


def test_find_out_of_sync_does_not_exist():
    (spec, name) = _find_out_of_sync_importable_spec([], "nopenopenope")
    assert spec is None
    assert name is None


def test_to_json():
    # the stuff from this parent env spec should NOT end up in the JSON
    hi = EnvSpec(name="hi",
                 conda_packages=['q', 'r'],
                 pip_packages=['zoo', 'boo'],
                 channels=['x1', 'y1'],
                 inherit_from_names=(),
                 inherit_from=())
    spec = EnvSpec(name="foo",
                   description="The Foo Spec",
                   conda_packages=['a', 'b'],
                   pip_packages=['c', 'd'],
                   channels=['x', 'y'],
                   inherit_from_names=('hi', ),
                   inherit_from=(hi, ))
    json = spec.to_json()

    assert {
        'description': "The Foo Spec",
        'channels': ['x', 'y'],
        'inherit_from': 'hi',
        'packages': ['a', 'b', {
            'pip': ['c', 'd']
        }]
    } == json


def test_to_json_no_description_no_pip_no_inherit():
    # should be able to jsonify a spec with no description
    spec = EnvSpec(name="foo",
                   conda_packages=['a', 'b'],
                   pip_packages=[],
                   channels=['x', 'y'],
                   inherit_from_names=(),
                   inherit_from=())
    json = spec.to_json()

    assert {'channels': ['x', 'y'], 'packages': ['a', 'b']} == json


def test_to_json_multiple_inheritance():
    spec = EnvSpec(name="foo",
                   conda_packages=['a', 'b'],
                   pip_packages=['c', 'd'],
                   channels=['x', 'y'],
                   inherit_from_names=('hi', 'hello'))
    json = spec.to_json()

    assert {
        'channels': ['x', 'y'],
        'inherit_from': ['hi', 'hello'],
        'packages': ['a', 'b', {
            'pip': ['c', 'd']
        }]
    } == json


def test_diff_from():
    spec1 = EnvSpec(name="foo", conda_packages=['a', 'b'], pip_packages=['c', 'd'], channels=['x', 'y'])
    spec2 = EnvSpec(name="bar", conda_packages=['a', 'b', 'q'], pip_packages=['c'], channels=['x', 'y', 'z'])
    diff = spec2.diff_from(spec1)

    assert '  channels:\n      x\n      y\n    + z\n  a\n  b\n+ q\n  pip:\n      c\n    - d' == diff


def test_save_environment_yml():
    def check_save(spec, dirname):
        saved = os.path.join(dirname, 'saved.yml')
        spec.save_environment_yml(saved)

        spec2 = _load_environment_yml(saved)

        assert spec2 is not None
        assert spec2.name == 'foo'
        assert spec2.conda_packages == ('xyz', 'bar=1.0', 'baz', 'abc')
        assert spec2.pip_packages == ('pippy', 'poppy==2.0')
        assert spec2.channels == ('channel1', 'channel2')

        assert spec2.logical_hash == 'ee1be9dc875857a69ccabb96cb45b5b828a6dff9'

    def check(filename):
        spec = _load_environment_yml(filename)

        assert spec is not None
        assert spec.name == 'foo'
        assert spec.conda_packages == ('xyz', 'bar=1.0', 'baz', 'abc')
        assert spec.pip_packages == ('pippy', 'poppy==2.0')
        assert spec.channels == ('channel1', 'channel2')

        assert spec.logical_hash == 'ee1be9dc875857a69ccabb96cb45b5b828a6dff9'

        with_directory_contents({}, lambda dirname: check_save(spec, dirname))

    with_file_contents(
        """
name: foo
dependencies:
  - xyz
  - bar=1.0
  - baz
  - abc
  - pip:
    - pippy
    - poppy==2.0
channels:
  - channel1
  - channel2
    """, check)


def test_overwrite_packages_with_lock_set():
    lock_set = CondaLockSet({'all': ['a=1.0=1']},
                            platforms=['linux-32', 'linux-64', 'osx-64', 'osx-arm64', 'win-32', 'win-64'])
    spec = EnvSpec(name="foo",
                   conda_packages=['a', 'b'],
                   pip_packages=['c', 'd'],
                   channels=['x', 'y'],
                   lock_set=lock_set)

    # package "b" is now ignored
    assert ('a=1.0=1', ) == spec.conda_packages_for_create


def test_lock_set_affects_name_sets():
    lock_set = CondaLockSet({'all': ['a=1.0=1', 'q=2.0=2']},
                            platforms=['linux-32', 'linux-64', 'osx-64', 'osx-arm64', 'win-32', 'win-64'])
    spec = EnvSpec(name="foo",
                   conda_packages=['a', 'b'],
                   pip_packages=['c', 'd'],
                   channels=['x', 'y'],
                   lock_set=lock_set)

    assert ('a', 'b') == spec.conda_packages
    assert ('a=1.0=1', 'q=2.0=2') == spec.conda_packages_for_create
    assert set(['a', 'b']) == spec.conda_package_names_set
    assert set(['a', 'q']) == spec.conda_package_names_for_create_set


def test_lock_set_affects_hash():
    lock_set = CondaLockSet({'all': ['a=1.0=1']},
                            platforms=['linux-32', 'linux-64', 'osx-64', 'osx-arm64', 'win-32', 'win-64'])
    with_lock_spec = EnvSpec(name="foo",
                             conda_packages=['a', 'b'],
                             pip_packages=['c', 'd'],
                             channels=['x', 'y'],
                             lock_set=lock_set)
    without_lock_spec = EnvSpec(name=with_lock_spec.name,
                                conda_packages=with_lock_spec.conda_packages,
                                pip_packages=with_lock_spec.pip_packages,
                                channels=with_lock_spec.channels,
                                lock_set=None)

    assert with_lock_spec.conda_packages != with_lock_spec.conda_packages_for_create
    assert without_lock_spec.conda_packages == without_lock_spec.conda_packages_for_create

    assert without_lock_spec.logical_hash == without_lock_spec.locked_hash
    assert with_lock_spec.logical_hash != with_lock_spec.locked_hash
    assert with_lock_spec.logical_hash == without_lock_spec.logical_hash

    assert without_lock_spec.locked_hash == without_lock_spec.import_hash
    assert with_lock_spec.locked_hash != with_lock_spec.import_hash


def test_platforms_affect_hash():
    with_platforms_spec = EnvSpec(name="foo",
                                  conda_packages=['a', 'b'],
                                  pip_packages=['c', 'd'],
                                  channels=['x', 'y'],
                                  platforms=('linux-64', ))
    without_platforms_spec = EnvSpec(name=with_platforms_spec.name,
                                     conda_packages=with_platforms_spec.conda_packages,
                                     pip_packages=with_platforms_spec.pip_packages,
                                     channels=with_platforms_spec.channels,
                                     platforms=())

    assert with_platforms_spec.logical_hash != with_platforms_spec.locked_hash
    assert with_platforms_spec.logical_hash != with_platforms_spec.import_hash

    assert without_platforms_spec.logical_hash == without_platforms_spec.locked_hash
    assert without_platforms_spec.logical_hash == without_platforms_spec.import_hash



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import pytest

from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.cli.prepare import prepare_command, main
from anaconda_project.internal.cli.prepare_with_mode import (UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT,
                                                             UI_MODE_TEXT_ASSUME_YES_PRODUCTION, UI_MODE_TEXT_ASSUME_NO)
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.local_state_file import LocalStateFile

from anaconda_project.test.project_utils import project_dir_disable_dedicated_env
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.internal import keyring


class Args(object):
    def __init__(self, **kwargs):
        self.directory = "."
        self.env_spec = None
        self.mode = UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT
        self.command = None
        for key in kwargs:
            setattr(self, key, kwargs[key])


def _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch):
    can_connect_args = dict()

    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        can_connect_args['host'] = host
        can_connect_args['port'] = port
        can_connect_args['timeout_seconds'] = timeout_seconds
        return True

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)

    return can_connect_args


def _test_prepare_command(monkeypatch, ui_mode):
    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def prepare_redis_url(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = prepare_command(dirname, ui_mode, conda_environment=None, command_name=None)
        assert can_connect_args['port'] == 6379
        assert result

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, prepare_redis_url)


def test_prepare_command_development(monkeypatch):
    _test_prepare_command(monkeypatch, UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT)


def test_prepare_command_production(monkeypatch):
    _test_prepare_command(monkeypatch, UI_MODE_TEXT_ASSUME_YES_PRODUCTION)


def test_prepare_command_assume_no(monkeypatch):
    _test_prepare_command(monkeypatch, UI_MODE_TEXT_ASSUME_NO)


def _monkeypatch_can_connect_to_socket_to_fail_to_find_redis(monkeypatch):
    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        if port == 6379:
            return False  # default Redis not there
        else:
            return True  # can't start a custom Redis here

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)


def test_main_fails_to_redis(monkeypatch, capsys):
    _monkeypatch_can_connect_to_socket_to_fail_to_find_redis(monkeypatch)

    from anaconda_project.internal.cli.prepare_with_mode import prepare_with_ui_mode_printing_errors as real_prepare

    def _mock_prepare_do_not_keep_going(project,
                                        environ=None,
                                        ui_mode=UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT,
                                        all=False,
                                        extra_command_args=None):
        return real_prepare(project, environ, ui_mode=ui_mode, extra_command_args=extra_command_args)

    monkeypatch.setattr('anaconda_project.internal.cli.prepare_with_mode.prepare_with_ui_mode_printing_errors',
                        _mock_prepare_do_not_keep_going)

    def main_redis_url(dirname):
        project_dir_disable_dedicated_env(dirname)
        code = main(Args(directory=dirname, all=False, refresh=False))
        assert 1 == code

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, main_redis_url)

    out, err = capsys.readouterr()
    assert "missing requirement" in err
    assert "All ports from 6380 to 6449 were in use" in err


def test_prepare_command_choose_environment(capsys, monkeypatch):
    def mock_conda_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        from anaconda_project.internal.makedirs import makedirs_ok_if_exists
        metadir = os.path.join(prefix, "conda-meta")
        makedirs_ok_if_exists(metadir)
        for p in pkgs:
            pkgmeta = os.path.join(metadir, "%s-0.1-pyNN.json" % p)
            open(pkgmeta, 'a').close()

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    def check_prepare_choose_environment(dirname):
        wrong_envdir = os.path.join(dirname, "envs", "foo")
        envdir = os.path.join(dirname, "envs", "bar")
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--env-spec=bar'])
        assert result == 0

        assert os.path.isdir(envdir)
        assert not os.path.isdir(wrong_envdir)

        package_json = os.path.join(envdir, "conda-meta", "nonexistent_bar-0.1-pyNN.json")
        assert os.path.isfile(package_json)

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME:
            """
env_specs:
  foo:
    packages:
        - nonexistent_foo
  bar:
    packages:
        - nonexistent_bar
"""
        }, check_prepare_choose_environment)

    out, err = capsys.readouterr()
    assert out == ("The project is ready to run commands.\n" +
                   "Use `anaconda-project list-commands` to see what's available.\n")
    assert err == ""


def test_prepare_command_all_environments(capsys, monkeypatch):
    def mock_conda_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        from anaconda_project.internal.makedirs import makedirs_ok_if_exists
        metadir = os.path.join(prefix, "conda-meta")
        makedirs_ok_if_exists(metadir)
        for p in pkgs:
            pkgmeta = os.path.join(metadir, "%s-0.1-pyNN.json" % p)
            open(pkgmeta, 'a').close()

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    def check_prepare_choose_environment(dirname):
        foo_envdir = os.path.join(dirname, "envs", "foo")
        bar_envdir = os.path.join(dirname, "envs", "bar")
        result = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname, '--all'])
        assert result == 0

        assert os.path.isdir(foo_envdir)
        assert os.path.isdir(bar_envdir)

        foo_package_json = os.path.join(foo_envdir, "conda-meta", "nonexistent_foo-0.1-pyNN.json")
        assert os.path.isfile(foo_package_json)

        bar_package_json = os.path.join(bar_envdir, "conda-meta", "nonexistent_bar-0.1-pyNN.json")
        assert os.path.isfile(bar_package_json)

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME:
            """
env_specs:
  foo:
    packages:
        - nonexistent_foo
  bar:
    packages:
        - nonexistent_bar
"""
        }, check_prepare_choose_environment)

    out, err = capsys.readouterr()
    assert out == ("The project is ready to run commands.\n" +
                   "Use `anaconda-project list-commands` to see what's available.\n")
    assert err == ""


def test_prepare_command_all_environments_refresh(capsys, monkeypatch):
    def mock_conda_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        from anaconda_project.internal.makedirs import makedirs_ok_if_exists
        metadir = os.path.join(prefix, "conda-meta")
        makedirs_ok_if_exists(metadir)
        for p in pkgs:
            pkgmeta = os.path.join(metadir, "%s-0.1-pyNN.json" % p)
            open(pkgmeta, 'a').close()

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    def check_prepare_choose_environment(dirname):
        foo_envdir = os.path.join(dirname, "envs", "foo")
        bar_envdir = os.path.join(dirname, "envs", "bar")
        result = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname, '--all'])
        assert result == 0

        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--all', '--refresh'])
        assert result == 0

        assert os.path.isdir(foo_envdir)
        assert os.path.isdir(bar_envdir)

        foo_package_json = os.path.join(foo_envdir, "conda-meta", "nonexistent_foo-0.1-pyNN.json")
        assert os.path.isfile(foo_package_json)

        bar_package_json = os.path.join(bar_envdir, "conda-meta", "nonexistent_bar-0.1-pyNN.json")
        assert os.path.isfile(bar_package_json)

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME:
            """
env_specs:
  foo:
    packages:
        - nonexistent_foo
  bar:
    packages:
        - nonexistent_bar
"""
        }, check_prepare_choose_environment)

    out, err = capsys.readouterr()
    assert out == ("The project is ready to run commands.\n" +
                   "Use `anaconda-project list-commands` to see what's available.\n" +
                   "The project is ready to run commands.\n" +
                   "Use `anaconda-project list-commands` to see what's available.\n")
    assert err == ""


def test_prepare_command_default_environment_refresh(capsys, monkeypatch):
    def mock_conda_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        from anaconda_project.internal.makedirs import makedirs_ok_if_exists
        metadir = os.path.join(prefix, "conda-meta")
        makedirs_ok_if_exists(metadir)
        for p in pkgs:
            pkgmeta = os.path.join(metadir, "%s-0.1-pyNN.json" % p)
            open(pkgmeta, 'a').close()

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    def check_prepare_choose_environment(dirname):
        default_envdir = os.path.join(dirname, "envs", "default")
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--env-spec', 'default'])
        assert result == 0

        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--env-spec', 'default', '--refresh'])
        assert result == 0

        assert os.path.isdir(default_envdir)

        foo_package_json = os.path.join(default_envdir, "conda-meta", "nonexistent_foo-0.1-pyNN.json")
        assert os.path.isfile(foo_package_json)

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: """
packages:
    - nonexistent_foo
"""}, check_prepare_choose_environment)

    out, err = capsys.readouterr()
    assert out == ("The project is ready to run commands.\n" +
                   "Use `anaconda-project list-commands` to see what's available.\n" +
                   "The project is ready to run commands.\n" +
                   "Use `anaconda-project list-commands` to see what's available.\n")
    assert err == ""


def test_prepare_command_choose_environment_does_not_exist(capsys):
    def check_prepare_choose_environment_does_not_exist(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--env-spec=nope'])
        assert result == 1

        expected_error = ("Environment name 'nope' is not in %s, these names were found: bar, foo" %
                          os.path.join(dirname, DEFAULT_PROJECT_FILENAME))
        out, err = capsys.readouterr()
        assert out == ""
        assert expected_error in err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME:
            """
env_specs:
  foo:
    packages:
        - nonexistent_foo
  bar:
    packages:
        - nonexistent_bar
"""
        }, check_prepare_choose_environment_does_not_exist)


@pytest.mark.slow
def test_prepare_command_choose_command_chooses_env_spec(capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--command=with_bar'])
        assert result == 1

        out, err = capsys.readouterr()
        assert 'nonexistent_bar' in err
        assert 'nonexistent_foo' not in err

        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'prepare', '--directory', dirname, '--command=with_foo'])
        assert result == 1

        out, err = capsys.readouterr()
        assert 'nonexistent_foo' in err
        assert 'nonexistent_bar' not in err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME:
            """
env_specs:
  foo:
    packages:
        - nonexistent_foo
  bar:
    packages:
        - nonexistent_bar
commands:
  with_foo:
    conda_app_entry: python --version
    env_spec: foo
  with_bar:
    conda_app_entry: python --version
    env_spec: bar

"""
        }, check)


def test_ask_variables_interactively(monkeypatch):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        inputs = ["foo", "bar"]

        def mock_console_input(prompt, encrypted):
            return inputs.pop(0)

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 0

        local_state = LocalStateFile.load_for_directory(dirname)
        assert local_state.get_value(['variables', 'FOO']) == 'foo'
        assert local_state.get_value(['variables', 'BAR_PASSWORD']) is None
        assert set(keyring.fallback_data().values()) == set(['bar'])

    keyring.enable_fallback_keyring()
    try:
        with_directory_contents_completing_project_file(
            {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
  BAR_PASSWORD: null
"""}, check)
    finally:
        keyring.disable_fallback_keyring()


def test_ask_variables_interactively_empty_answer_re_asks(monkeypatch):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        inputs = ["", "foo", "bar"]

        def mock_console_input(prompt, encrypted):
            return inputs.pop(0)

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 0

        local_state = LocalStateFile.load_for_directory(dirname)
        assert local_state.get_value(['variables', 'FOO']) == 'foo'
        assert local_state.get_value(['variables', 'BAR']) == 'bar'

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
  BAR: null
"""}, check)


def test_ask_variables_interactively_whitespace_answer_re_asks(monkeypatch):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        inputs = ["    ", "foo", "bar"]

        def mock_console_input(prompt, encrypted):
            return inputs.pop(0)

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 0

        local_state = LocalStateFile.load_for_directory(dirname)
        assert local_state.get_value(['variables', 'FOO']) == 'foo'
        assert local_state.get_value(['variables', 'BAR']) == 'bar'

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
  BAR: null
"""}, check)


_foo_and_bar_missing = ("missing requirement to run this project: BAR environment variable must be set.\n" +
                        "  Environment variable BAR is not set.\n" +
                        "missing requirement to run this project: FOO environment variable must be set.\n" +
                        "  Environment variable FOO is not set.\n")


def test_ask_variables_interactively_eof_answer_gives_up(monkeypatch, capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt, encrypted):
            return None

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 1

        out, err = capsys.readouterr()

        assert err == _foo_and_bar_missing

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
  BAR: null
"""}, check)


def test_ask_variables_interactively_then_set_variable_fails(monkeypatch, capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        inputs = ["foo", "bar"]

        def mock_console_input(prompt, encrypted):
            return inputs.pop(0)

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        def mock_set_variables(project, env_spec_name, vars_and_values, prepare_result):
            return SimpleStatus(success=False, description="Set variables FAIL")

        monkeypatch.setattr('anaconda_project.project_ops.set_variables', mock_set_variables)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 1

        out, err = capsys.readouterr()

        assert err == _foo_and_bar_missing + "Set variables FAIL\n"

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
  BAR: null
"""}, check)


def test_no_ask_variables_interactively_not_interactive(monkeypatch, capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return False

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt, encrypted):
            raise Exception("should not have been called")

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 1

        out, err = capsys.readouterr()

        assert err == _foo_and_bar_missing

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
  BAR: null
"""}, check)


def test_no_ask_variables_interactively_if_no_variables_missing_but_prepare_fails(monkeypatch, capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt, encrypted):
            raise Exception("Should not have called this, prompt " + prompt)

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 1

        out, err = capsys.readouterr()

        assert out == ""
        assert err == ("%s: env_specs should be a dictionary from environment name to environment attributes, not 42\n"
                       "Unable to load the project.\n") % DEFAULT_PROJECT_FILENAME

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME:
            """
variables:
  FOO: { default: "foo" }
  BAR: { default: "bar" }

# breakage
env_specs: 42

"""
        }, check)


@pytest.mark.slow
def test_no_ask_conda_prefix_interactively(monkeypatch, capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt, encrypted):
            raise Exception("should not have been called")

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 1

        out, err = capsys.readouterr()

        assert err.endswith("Conda environment is missing packages: nonexistent_package_name\n")

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
packages:
 - nonexistent_package_name
"""}, check)


def test_display_suggestions(monkeypatch, capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)

        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt, encrypted):
            raise Exception("should not have been called")

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        res = _parse_args_and_run_subcommand(['anaconda-project', 'prepare', '--directory', dirname])
        assert res == 0

        out, err = capsys.readouterr()

        assert """Potential issues with this project:
  * anaconda-project.yml: Unknown field name 'weird_field'

The project is ready to run commands.
Use `anaconda-project list-commands` to see what's available.
""" == out
        assert '' == err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: """
packages: []
weird_field: 42
"""}, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Project "local state" file loading and manipulation."""
from __future__ import absolute_import

import os

from anaconda_project.yaml_file import YamlFile

# these are in the order we'll use them if multiple are present
possible_local_state_file_names = ("anaconda-project-local.yml", "anaconda-project-local.yaml")

DEFAULT_LOCAL_STATE_FILENAME = possible_local_state_file_names[0]

SERVICE_RUN_STATES_SECTION = "service_run_states"


class LocalStateFile(YamlFile):
    """Represents the locally-configured/user-specific state of the project directory.

    Project config that you might want in source control should be
    in ``ProjectFile`` instead.

    Be careful with creating your own instance of this class,
    because you have to think about when other code might load or
    save in a way that conflicts with your loads and saves.
    """

    # The dummy entry works around a bug/quirk in ruamel.yaml that drops the
    # top comment for an empty dictionary
    template = '''
# Anaconda local project state (specific to this user/machine)
__dummy__: dummy
'''

    @classmethod
    def load_for_directory(cls, directory, scan_parents=True):
        """Load the project local state file from the given directory, even if it doesn't exist.

        If the directory has no project file, the loaded
        ``LocalStateFile`` will be empty. It won't actually be
        created on disk unless you call ``save()``.

        If the local state file has syntax problems, the
        ``corrupted`` and ``corrupted_error_message`` properties
        will be set and attempts to modify or save the file will
        raise an exception.

        Args:
            directory (str): path to the project directory

        Returns:
            a new ``LocalStateFile``

        """
        current_dir = directory
        while current_dir != os.path.realpath(os.path.dirname(current_dir)):
            for name in possible_local_state_file_names:
                path = os.path.join(current_dir, name)
                if os.path.isfile(path):
                    return LocalStateFile(path)

            if scan_parents:
                current_dir = os.path.dirname(os.path.abspath(current_dir))
                continue
            else:
                break

        # No file was found, create a new one
        return LocalStateFile(os.path.join(directory, DEFAULT_LOCAL_STATE_FILENAME))

    def set_service_run_state(self, service_name, state):
        """Set a dict value in the ``service_run_states`` section.

        This is used to save the state of a running service, such as
        its port or process ID.

        Conventionally the state can also include a
        ``shutdown_commands`` property with a list of list of
        strings value. The lists of strings are args to pass to
        exec. In order to shut down a service, we run all
        ``shutdown_commands`` and then delete the entire service
        run state dict.

        This method does not save the file, call ``save()`` to do that.

        Args:
            service_name (str): environment variable identifying the service
            state (dict): state for the running service process
        """
        if not isinstance(state, dict):
            raise ValueError("service state should be a dict")
        self.set_value([SERVICE_RUN_STATES_SECTION, service_name], state)

    def get_service_run_state(self, service_name):
        """Get the running instance state for a service.

        Args:
            service_name (str): environment variable identifying the service

        Returns:
            The state dict (empty dict if no state was saved)
        """
        return self.get_value([SERVICE_RUN_STATES_SECTION, service_name], default=dict())

    def get_all_service_run_states(self):
        """Get all saved service run states.

        Returns the entire ``service_run_states`` section.

        Returns:
            a dict from service name to service state dict
        """
        return self.get_value(SERVICE_RUN_STATES_SECTION, default=dict())



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from anaconda_project.internal.test.fake_frontend import FakeFrontend


def test_partial_messages():
    frontend = FakeFrontend()

    frontend.partial_info("a\nb\nc\nd")
    # d is stuck in the buffer
    assert frontend.logs == ['a', 'b', 'c']
    assert frontend._info_buf == 'd'

    frontend.partial_error("1")
    frontend.partial_error("2")
    frontend.partial_error("3")
    frontend.partial_error("456")
    assert frontend._error_buf == "123456"
    frontend.partial_error("\n")

    assert frontend._error_buf == ""
    assert frontend.errors == ['123456']

    frontend.partial_error("\n\n\n\n")
    assert frontend.errors == ['123456', '', '', '', '']


def test_partial_messages_with_windows_line_endings():
    frontend = FakeFrontend()

    frontend.partial_info("a\r\nb\r\nc\r\nd")
    # d is stuck in the buffer
    assert frontend.logs == ['a', 'b', 'c']
    assert frontend._info_buf == 'd'



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Types related to project requirements."""
from __future__ import absolute_import

from abc import ABCMeta, abstractmethod
from copy import deepcopy

from anaconda_project.internal.metaclass import with_metaclass
from anaconda_project.internal.py2_compat import is_string
from anaconda_project.status import Status


class UserConfigOverrides(object):
    """Class containing user-forced configuration for the prepare process."""
    def __init__(self, inherited_env=None, env_spec_name=None):
        """Construct a set of user overrides for the prepare process."""
        self._inherited_env = inherited_env
        self._env_spec_name = env_spec_name

    @property
    def env_spec_name(self):
        """The user-specified name of the conda environment spec to use, or None if not specified."""
        return self._env_spec_name

    @property
    def inherited_env(self):
        """The environment we started with before we ran the prepare process."""
        return self._inherited_env

    @env_spec_name.setter
    def env_spec_name(self, value):
        """Change the conda environment name override."""
        self._env_spec_name = value


class RequirementStatus(Status):
    """Class describing the status of a requirement.

    Values of this class are immutable; to get updated status, you
    would call ``recheck()`` to get a new status.

    """
    def __init__(self, requirement, has_been_provided, status_description, provider, analysis, latest_provide_result,
                 env_spec_name):
        """Construct an abstract RequirementStatus."""
        self._requirement = requirement
        self._has_been_provided = has_been_provided
        self._status_description = status_description
        self._provider = provider
        self._analysis = analysis
        self._latest_provide_result = latest_provide_result
        self._env_spec_name = env_spec_name

    def __repr__(self):
        """Repr of the status."""
        return "RequirementStatus(%r,%r,%r)" % (self.has_been_provided, self.status_description, self.requirement)

    def __bool__(self):
        """True if the requirement is met."""
        return self.has_been_provided

    def __nonzero__(self):
        """True if the requirement is met."""
        return self.__bool__()  # pragma: no cover (py2 only)

    @property
    def requirement(self):
        """Get the requirement we are the status of."""
        return self._requirement

    @property
    def has_been_provided(self):
        """Get True if the requirement has been met."""
        return self._has_been_provided

    @property
    def status_description(self):
        """Get a description of how the requirement has been met (or why it hasn't been)."""
        return self._status_description

    @property
    def provider(self):
        """Get the provider for this requirement."""
        return self._provider

    @property
    def analysis(self):
        """Get the provider's analysis of the status."""
        return self._analysis

    @property
    def latest_provide_result(self):
        """Get the latest ``ProvideResult`` or None if we haven't provided yet."""
        return self._latest_provide_result

    @property
    def errors(self):
        """Get error logs relevant to the status, from either checking status or attempting to provide it."""
        if self.latest_provide_result is None:
            return []
        else:
            return self.latest_provide_result.errors

    @property
    def env_spec_name(self):
        """Get the env spec name used to meet the requirement, None if not a ``CondaEnvRequirement``."""
        return self._env_spec_name

    def recheck(self, environ, local_state_file, default_env_spec_name, overrides=None, latest_provide_result=None):
        """Get a new ``RequirementStatus`` reflecting the current state.

        This calls ``Requirement.check_status()`` which can do network and filesystem IO,
        so be cautious about where you call it.
        """
        if latest_provide_result is None:
            latest_provide_result = self._latest_provide_result
        return self.requirement.check_status(environ, local_state_file, default_env_spec_name, overrides,
                                             latest_provide_result)


class Requirement(with_metaclass(ABCMeta)):
    """Describes a requirement of the project (from the project config).

    Note that this is not specifically a package requirement;
    this class is more general, it can be a requirement for any
    condition at all (that a service is running, that a file
    exists, or even that a package is intalled - anything you can
    think of).

    """
    def __init__(self, registry, options):
        """Construct a Requirement.

        Args:
            registry (RequirementsRegistry): the plugin registry we came from
            options (dict): dict of requirement options from the project config
        """
        self.registry = registry

        if options is None:
            self.options = dict()
        else:
            self.options = deepcopy(options)
            # always convert the default to a string (it's allowed to be a number
            # in the config file, but env vars have to be strings), unless
            # it's a dict because we use a dict for encrypted defaults
            if 'default' in self.options and not (is_string(self.options['default'])
                                                  or isinstance(self.options['default'], dict)):
                self.options['default'] = str(self.options['default'])

    @property
    @abstractmethod
    def title(self):
        """Human-readable short name of the requirement."""
        pass  # pragma: no cover

    def _description(self, default):
        """Use this in subclasses to implement the description property."""
        if 'description' in self.options:
            return self.options['description']
        else:
            return default

    @property
    @abstractmethod
    def description(self):
        """Human-readable about-one-sentence hint or tooltip for the requirement."""
        pass  # pragma: no cover

    @property
    def ignore_patterns(self):
        """Set of ignore patterns for files this requirement's provider might autogenerate."""
        return set()

    def _create_status(self, environ, local_state_file, default_env_spec_name, overrides, latest_provide_result,
                       has_been_provided, status_description, provider_class_name):
        provider = self.registry.find_provider_by_class_name(provider_class_name)
        analysis = provider.analyze(self, environ, local_state_file, default_env_spec_name, overrides)
        env_spec_name = analysis.config.get('env_name', None)
        return RequirementStatus(self,
                                 has_been_provided=has_been_provided,
                                 status_description=status_description,
                                 provider=provider,
                                 analysis=analysis,
                                 latest_provide_result=latest_provide_result,
                                 env_spec_name=env_spec_name)

    def _create_status_from_analysis(self, environ, local_state_file, default_env_spec_name, overrides,
                                     latest_provide_result, provider_class_name, status_getter):
        provider = self.registry.find_provider_by_class_name(provider_class_name)
        analysis = provider.analyze(self, environ, local_state_file, default_env_spec_name, overrides)
        (has_been_provided, status_description) = status_getter(environ, local_state_file, analysis)
        env_spec_name = analysis.config.get('env_name', None)

        return RequirementStatus(self,
                                 has_been_provided=has_been_provided,
                                 status_description=status_description,
                                 provider=provider,
                                 analysis=analysis,
                                 latest_provide_result=latest_provide_result,
                                 env_spec_name=env_spec_name)

    @abstractmethod
    def check_status(self, environ, local_state_file, default_env_spec_name, overrides, latest_provide_result=None):
        """Check on the requirement and return a ``RequirementStatus`` with the current status.

        This may attempt to talk to servers, check that files
        exist on disk, and other work of that nature to verify the
        requirement's status, so be careful about when and how
        often this gets called.

        Args:
            environ (dict): use this rather than the system environment directly
            local_state_file (LocalStateFile): project local state
            default_env_spec_name (str): fallback env spec name
            overrides (UserConfigOverrides): user-supplied forced choices
            latest_provide_result (ProvideResult): most recent result of ``provide()`` or None

        Returns:
            a ``RequirementStatus``

        """
        pass  # pragma: no cover (abstract method)


# suffixes that change the default for the "encrypted" option
_secret_suffixes = ('_PASSWORD', '_SECRET_KEY', '_SECRET')


class EnvVarRequirement(Requirement):
    """A requirement that a certain environment variable be set."""
    @classmethod
    def _parse_default(self, options, env_var, problems):
        assert (isinstance(options, dict))

        raw_default = options.get('default', None)

        if raw_default is None:
            good_default = True
        elif isinstance(raw_default, bool):
            # we have to check bool since it's considered an int apparently
            good_default = False
        elif is_string(raw_default) or isinstance(raw_default, (int, float)):
            good_default = True
        else:
            good_default = False

        if 'default' in options and raw_default is None:
            # convert null to be the same as simply missing
            del options['default']

        if good_default:
            return True
        else:
            problems.append(
                "default value for variable {env_var} must be null, a string, or a number, not {value}.".format(
                    env_var=env_var, value=raw_default))
            return False

    def __init__(self, registry, env_var, options=None):
        """Construct an EnvVarRequirement for the given ``env_var`` with the given options."""
        super(EnvVarRequirement, self).__init__(registry, options)
        assert env_var is not None
        self.env_var = env_var

    def __repr__(self):
        """Custom repr of EnvVarRequirement."""
        return "%s(env_var='%s')" % (self.__class__.__name__, self.env_var)

    @property
    def title(self):
        """Override superclass title."""
        return self.env_var

    @property
    def description(self):
        """Override superclass description."""
        return self._description("%s environment variable must be set." % (self.env_var))

    @property
    def encrypted(self):
        """Get whether this is a password-type value we should encrypt when possible."""
        if 'encrypted' in self.options:
            return self.options['encrypted']
        else:
            return any(self.env_var.endswith(suffix) for suffix in _secret_suffixes)

    @property
    def default_as_string(self):
        """Get the default, forced to string or None."""
        value = self.options.get('default', None)
        if value is None:
            return None
        else:
            # see _parse_default above, it can be a string already,
            # or an integer
            return str(value)

    def _get_value_of_env_var(self, environ):
        """A "protected" method for subtypes to use."""
        value = environ.get(self.env_var, None)
        if value == "":  # do we really want this, maybe empty string is a valid value sometimes?
            value = None
        return value

    def _unset_message(self):
        """A "protected" method for subtypes to use."""
        return "Environment variable {env_var} is not set.".format(env_var=self.env_var)

    def _set_message(self, environ):
        """A "protected" method for subtypes to use."""
        if self.encrypted:
            # don't include the value if it's an encrypted variable.
            return "Environment variable {env_var} is set.".format(env_var=self.env_var)
        else:
            return "Environment variable {env_var} set to '{value}'".format(env_var=self.env_var,
                                                                            value=self._get_value_of_env_var(environ))

    def check_status(self, environ, local_state_file, default_env_spec_name, overrides, latest_provide_result=None):
        """Override superclass to get our status."""
        value = self._get_value_of_env_var(environ)

        has_been_provided = value is not None
        if has_been_provided:
            status_description = self._set_message(environ)
        else:
            status_description = self._unset_message()

        return self._create_status(environ,
                                   local_state_file,
                                   default_env_spec_name=default_env_spec_name,
                                   overrides=overrides,
                                   has_been_provided=has_been_provided,
                                   status_description=status_description,
                                   provider_class_name='EnvVarProvider',
                                   latest_provide_result=latest_provide_result)
